// Copyright (C) 2003 Stephane Thiell
//
// This file is part of pxyscand (from pxys)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
#define RCSID "$Id: PXMSocks.cc,v 1.2 2004/01/01 02:18:10 mbuna Exp $"

#define SOCKS4_SHORTNAME "Socks4"
#define SOCKS5_SHORTNAME "Socks5"

#define SOCKS4_DESCR     "Insecure Socks 4 server"
#define SOCKS5_DESCR     "Insecure Socks 5 server"

#define SOCKS_PORT       1080

#define MAX_LINES_TRY 4

#include "PXMSocks.h"
#include "socks4.h"
#include "socks5.h"

#include <cassert>
#include <cerrno>
#include <cstring>

uint32_t PXMSocks::sConn4Count = 0;
uint32_t PXMSocks::sProxy4Count = 0;
uint32_t PXMSocks::sConn5Count = 0;
uint32_t PXMSocks::sProxy5Count = 0;

PXMSocks::PXMSocks(PXScan *inScan)
  : PXScanModule(inScan), mStatus(SCAN_NOSOCKS), mLinesTry(0)
  {
  }

PXMSocks::~PXMSocks()
  {
  }

void
PXMSocks::InitModule()
  {
  RegisterPXM(SOCKS4_SHORTNAME, SOCKS_PORT, &sConn4Count, &sProxy4Count);
  RegisterPXM(SOCKS5_SHORTNAME, SOCKS_PORT, &sConn5Count, &sProxy5Count);
  }

bool
PXMSocks::StartScan()
  {
  peak_task task = peak_task_self();
  
  struct sockaddr_in sin;
  memset(&sin, 0, sizeof(struct sockaddr_in));
  sin.sin_family = AF_INET;
  sin.sin_addr = this->GetAddress();
  sin.sin_port = htons(SOCKS_PORT);
  
  mStream = peak_stream_socket_create((struct sockaddr *)&sin, sizeof(sin),
              mStatus == SCAN_NOSOCKS ? PEAK_STREAM_OPT_LINEMODE
                                      : PEAK_STREAM_OPT_DEFAULT,
              EventCallback,
              this);
  
  if (!mStream)
    return false;
  
  if (this->IsLocalAddressSet())
    {
    sockaddr_in local_sin;
    memset(&local_sin, 0, sizeof(local_sin));
    local_sin.sin_family = AF_INET;
    local_sin.sin_addr = this->GetLocalAddress();
    local_sin.sin_port = htons(0);
    
    peak_stream_set_address(mStream, (sockaddr*)&local_sin, sizeof(local_sin));
    }
  
  /* Enable built-in timeout option, this is so useful here. */
  peak_stream_set_timeout(mStream, GetTimeout());
  
  /* Connect (don't block) */
  if (peak_stream_connect(mStream) == -1)
    {
    this->Cleanup();
    this->ProxyNotFound();
    }
  else
    peak_stream_schedule(mStream, task);
  return true;
  }

void
PXMSocks::Cleanup()
  {
  peak_release(mStream);
  }

void
PXMSocks::ProxyNotFound()
  {
  if (mStatus == SCAN_SOCKS4)
    {
    mStatus = SCAN_SOCKS5;
    this->StartScan();
    return;
    }
  
  PXScanModule::ProxyNotFound();
  }

void
PXMSocks::SendSocks4Request(peak_stream s)
  {
  struct socks4 socks4out = { S4_VN, S4_CMD_CONNECT, { 0x1A, 0x0B },
                              { 0, 0, 0, 0 }, 0 };
  in_addr targetAddr = GetTargetAddress();
  
  /* Set target address */
  memcpy(&socks4out.dstip, &targetAddr, sizeof(socks4out.dstip));
  
  peak_stream_set_buffered(s, 1, sizeof(socks4out), sizeof(socks4out), NULL);
  peak_stream_write_buffered(s, &socks4out, sizeof(socks4out));
  }

void
PXMSocks::SendSocks5Request(peak_stream s)
  {
  struct socks5 socks5out = { S5_VN, S5_CMD_CONNECT };
  
  peak_stream_set_buffered(s, 1, sizeof(socks5out), sizeof(socks5out), NULL);
  peak_stream_write_buffered(s, &socks5out, sizeof(socks5out));
  }

void
PXMSocks::ProcessEvent(peak_stream s, int type)
  {
  int err;
  
  switch (type)
    {
    case PEAK_STREAM_EVT_OPEN:
      if (mStatus == SCAN_NOSOCKS)
        {
        mStatus = SCAN_SOCKS4;
        sConn4Count++;
        this->SendSocks4Request(s);
        }
      else
        {
        assert(mStatus == SCAN_SOCKS5);
        sConn5Count++;
        this->SendSocks5Request(s);
        }
      break;
    case PEAK_STREAM_EVT_READ:
      if (mStatus == SCAN_SOCKS4)
        {
        char *line = peak_stream_get_line(s);
        if (this->ScannerStringCheck(line))
          {
          sProxy4Count++;
          this->Cleanup();
          this->ProxyFound(OPAS_PROXY_TYPE_SOCKS4, SOCKS_PORT,
                           SOCKS4_DESCR);
          return; /* done! */
          }
        if (mLinesTry++ < MAX_LINES_TRY)
          break;
        }
      else
        {
        // Socks v5
        struct socks5_response socks5in;
        if (peak_stream_read(s, &socks5in, sizeof(socks5in)) > 1)
          {
          if (socks5in.vn == S5_VN)
            {
            switch(socks5in.method)
              {
              case S5_R_METHOD_NOAUTH:
                sProxy5Count++;
                this->Cleanup();
                this->ProxyFound(OPAS_PROXY_TYPE_SOCKS5, SOCKS_PORT,
                                 SOCKS5_DESCR);
                return; /* done! */
              case S5_R_METHOD_GSSAPI:
              case S5_R_METHOD_USERPW:
              case S5_R_METHOD_NOTACC:
              default:
                break;
              }
            }
          }
        }
      /* fall through */
    case PEAK_STREAM_EVT_ERROR:
    case PEAK_STREAM_EVT_TIMEDOUT:
      this->Cleanup();
      this->ProxyNotFound();
      break;
    case PEAK_STREAM_EVT_END:
      err = peak_stream_get_error(s);
      this->Cleanup();
      if (err == ENETUNREACH)
        this->ScanError(OPAS_ERROR_NETUNREACH); // Can't scan !
      else if (err == ENETDOWN)
        this->ScanError(OPAS_ERROR_NETDOWN);    // Even worst !
      else
        this->ProxyNotFound();
      break;
    default:
      break;
    }
  }

void
PXMSocks::EventCallback(peak_stream s, int type, void *context)
  {
  PXMSocks *pxm = reinterpret_cast<PXMSocks*>(context);
  pxm->ProcessEvent(s, type);
  }
